package com.cloveretl.consulting.cryptography.bouncycastle;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.util.regex.Pattern;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.ShortBufferException;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import org.jetel.component.AbstractGenericTransform;
import org.jetel.data.DataRecord;
import org.jetel.exception.ComponentNotReadyException;
import org.jetel.exception.ConfigurationStatus;
import org.jetel.exception.ConfigurationStatus.Priority;
import org.jetel.exception.ConfigurationStatus.Severity;
import org.jetel.metadata.DataRecordMetadata;

/**
 * THIS CLASS IS NEITHER MEANT TO BE USED IN PRODUCTION ENVIRONMENT NOR DOES SET STANDARD FOR FULLY OPTIMIZED
 * CODE. THIS PIECE OF CODE HAS NOT BEEN TESTED EXTENSIVELY AND TRIES TO BALANCE BETWEEN
 * COMPREHENCIBILITY AND COMPLEXICITY, THUS DOES NOT COVER ALL BORDER CONDITIONS!
 * 
 * This class represents example, how CloverETL may utilize features brought by external Java libraries.
 * Large portion of a code (classes implementing IEncryptionApi) was taken and re-purposed from Bouncy Castle Java examples:
 * 	http://media.wiley.com/product_ancillary/30/07645963/DOWNLOAD/beg_crypto_examples.zip 
 * All credits for this code goes to its respective owner.
 * 
 * Class is designed to be used as an "engine" of CustomJavaTransform component of an CloverETL software,
 * it utilizes Java library Bouncy Castle (https://www.bouncycastle.org/java.html). Based on configured target field
 * data type, it outputs either hexadecimal or binary representation of encrypted incoming string. Encryption algorithm
 * can be either AES (block) or ARC4 (stream). Code may be used for both encryption and decryption of an incoming data.
 * 
 * @author Pavel Svec <pavel.svec@javlin.eu>
 */
public class EncryptorDecryptor extends AbstractGenericTransform {
	/*
	 * List of accepted attributes, it is a good practice to have them defined here as constants
	 * so they can be easily changed and linked to the code.
	 */
	final static String PROP_ALGORITHM = "EncryptionAlgorithm";
	final static String PROP_MODE = "EncryptionMode";
	final static String PROP_FIELD = "EncryptedField";
	final static String PROP_TARGET = "EncryptionTarget";
	final static String PROP_VECTOR = "InitializationVector";
	final static String PROP_KEY = "EncryptionKey";
	/*
	 * Enumeration of supported ciphers
	 */
	public enum CipherType {
		AES,ARC4
	}
	/*
	 * Enumeration of available cipher operations
	 */
	public enum CipherMode {
		ENCRYPT,DECRYPT
	}
	/*
	 * Global variables used through several stages of live of an component
	 */
	IEncryptionApi encryption;		// Instance of an encryption algorithm wrapper
	Exception initException;		// Exception caught during component initiation procedure
	int fieldIdx;					// Index of field used for encryption - good practice is to find index of
									// such field and use it rather than refer to a field by its name
									// which has severe performance impact
	boolean sourceBinary;			// Data type of a source field
	int targetIdx;					// Same thing as for fieldIdx, but for a target field
	boolean targetBinary;			// Data type of an target field

	@Override
	public void execute() throws Exception {

		if (initException != null) {
			throw initException;
		}

		/*
		 * For each connected input port, there is an object which is being
		 * REUSED with arrival of any new record. CloverETL uses this approach
		 * to reduce overhead of making new object every time new record
		 * is read/created. Data flowing through components is basically a
		 * binary data stream.
		 */
		/** Record prepared for reading from input port 0 */
		DataRecord inRecord = inRecords[0];

		/** Record prepared for writing to output port 0 */
		DataRecord outRecord = outRecords[0];
		DataRecord errorRecord = outRecords[1];

		/*
		 * Read input port 0, until there is nothing to read
		 */
		long counter = 0;
		while ((inRecord = readRecordFromPort(0)) != null) {
			/*
			 * Take value from field user specified and convert its value to a String
			 * Subgraph configuration should restrict metadata field to string-only data type (see ENCRYPTED_FIELD parameter configuration)
			 */
			byte[] encValue;
			
			if (sourceBinary)
				encValue = (byte[]) inRecord.getField(fieldIdx).getValue();
			else
				encValue = inRecord.getField(fieldIdx).getValue().toString().getBytes();
			
			try {
				byte[] encrypted = encryption.encrypt(encValue);
				// Copy record from input to output
				outRecord.copyFieldsByName(inRecord);
				// Use encryption wrapper to encrypt incoming value and rewrite fieldIdx with that value
				// Output of an encryption algorithm is a hexadecimal representation of encrypted/decrypted string
				if (targetBinary) {
					outRecord.getField(targetIdx).setValue(encrypted);
				} else if (! sourceBinary) {
					outRecord.getField(targetIdx).setValue(toHex(encrypted));
				} else {
					outRecord.getField(targetIdx).setValue(new String(encrypted));
				}
				// Write out record to output
				writeRecordToPort(0, outRecord);
				/*
				 *  Good practice is to reset output record, otherwise it may be possible for some output
				 *  records to be populated by values from previous one. Not necessary in this case,
				 *  since we're copying everything from input to output.
				 */
				//outRecord.reset();
			} catch (Exception ex) {
				/*
				 * If something goes wrong with encryption, output using error port
				 */
				errorRecord.getField(0).setValue(counter);
				errorRecord.getField(1).setValue(ex.getMessage());
				errorRecord.getField(2).setValue(encValue);
				writeRecordToPort(1, errorRecord);
			}
			
			counter++;
		}
	}

	@Override
	public ConfigurationStatus checkConfig(ConfigurationStatus status) {
		super.checkConfig(status);
		
		String algorithm = getProperties().getProperty(PROP_ALGORITHM,"AES");
		String mode = getProperties().getProperty(PROP_MODE,"ENCRYPT");
		String fieldName = getProperties().getProperty(PROP_FIELD);
		String targetName = getProperties().getProperty(PROP_TARGET);
		String encKey = getProperties().getProperty(PROP_KEY);
		String iVector = getProperties().getProperty(PROP_VECTOR);
		
		DataRecordMetadata inMeta = getComponent().getInMetadata().get(0);
		DataRecordMetadata outMeta = getComponent().getOutMetadata().get(0);
		
		Pattern re = Pattern.compile("^(?:[0-9a-zA-Z])+$");
		
		// Validate input and output metadata records (only the ones being injected from master graph)
		checkMetadataRecord(fieldName,inMeta,status);
		checkMetadataRecord(targetName,outMeta,status);
		
		// Check for valid algorithm
		if (CipherType.valueOf(algorithm) == null)
			status.add("Invalid encryption algorithm '"+ algorithm +"', choose one of "+ CipherType.values() +".", Severity.ERROR, getComponent(), Priority.NORMAL);
		
		// Check for valid operation mode
		if (CipherMode.valueOf(mode) == null)
			status.add("Invalid encryption mode '"+ mode +"', choose one of "+ CipherMode.values() +".", Severity.ERROR, getComponent(), Priority.NORMAL);
		
		// Check for valid encryption key
		if (encKey == null || encKey.length() != 32 || !re.matcher(encKey).find())
			status.add("Encryption key is supposed to be 128-bit (32 characters long) hexadecimal string.", Severity.ERROR, getComponent(), Priority.NORMAL);
		
		// Check for initial vector
		if ((iVector == null || iVector.length() != 32 || !re.matcher(iVector).find()) && CipherType.valueOf(algorithm) == CipherType.AES)
			status.add("Initial vector is supposed to be 128-bit (32 characters long) hexadecimal string.", Severity.ERROR, getComponent(), Priority.NORMAL);
		
		return status;
	}
	
	/**
	 * Check metadata record for chosen field name and its validity - has to be string or byte
	 * enriches object "status" for all errors it may find
	 * 
	 * @param fieldName
	 * @param metaRecord
	 * @param status
	 */
	private void checkMetadataRecord(String fieldName, DataRecordMetadata metaRecord, ConfigurationStatus status) {
		String dtype;
		int fieldIdx;
		// Parameter set?
		if (fieldName != null) {
			// If yes, does field exist in metadata record? Return value -1 means, field was not found.
			if ((fieldIdx = metaRecord.getFieldPosition(fieldName)) < 0)
				status.add(String.format("Column '%s' does not exist in incoming metadata.",fieldName), Severity.WARNING, getComponent(), Priority.NORMAL);
			
			// If parameter is set and exists, is it one of the allowed types?
			else if (!(dtype = metaRecord.getField(fieldIdx).getDataType().getName()).equals("string") && !dtype.equals("byte"))
				status.add(String.format("Column '%s' is neither 'string' nor 'byte'.",fieldName), Severity.WARNING, getComponent(), Priority.NORMAL);
		
		// Existence of attribute is going to be checked on subgraph level (parameter is required) - there is no need to check it in here.
		} /* else {
			status.add("Empty field name attribute.", Severity.ERROR, getComponent(), Priority.NORMAL);
		} */
	}

	@Override
	public void init() {
		super.init();
		
		/*
		 * Read user-defined values from component configuration
		 */
		String algorithm = getProperties().getProperty(PROP_ALGORITHM,"AES");
		String mode = getProperties().getProperty(PROP_MODE,"ENCRYPT");
		String fieldName = getProperties().getProperty(PROP_FIELD);
		String targetName = getProperties().getProperty(PROP_TARGET);
		String encKey = getProperties().getProperty(PROP_KEY);
		String iVector = getProperties().getProperty(PROP_VECTOR);
		/*
		 * Output information read from configuration to a system log
		 */
		getLogger().info(String.format("Parameter [%s] resolved as: %s", PROP_ALGORITHM, algorithm));
		getLogger().info(String.format("Parameter [%s] resolved as: %s", PROP_MODE, mode));
		getLogger().info(String.format("Parameter [%s] resolved as: %s", PROP_FIELD, fieldName));
		getLogger().info(String.format("Parameter [%s] resolved as: %s", PROP_KEY, encKey));
		getLogger().info(String.format("Parameter [%s] resolved as: %s", PROP_VECTOR, iVector));
		
		try {
			switch (CipherType.valueOf(algorithm.toUpperCase())) {
			case AES:
				encryption = new AesEncryptor(fromHex(iVector),fromHex(encKey));
			break;
			case ARC4:
				encryption = new Arc4Encryptor(fromHex(encKey));
				break;
			}
			
			encryption.setMode(CipherMode.valueOf(mode));
			
		} catch (InvalidKeyException | NoSuchAlgorithmException
				| NoSuchProviderException | NoSuchPaddingException | InvalidAlgorithmParameterException e) {
			initException = e;
			return;
		}
		
		// Finds and stores index of an altered column
		fieldIdx = getComponent().getInMetadata().get(0).getFieldPosition(fieldName);
		targetIdx = getComponent().getOutMetadata().get(0).getFieldPosition(targetName);
		
		if (fieldIdx < 0 || targetIdx < 0) {
			initException = new IllegalStateException("Invalid source or target field. Please check your configuration.");
			return;
		}
		
		sourceBinary = getComponent().getInMetadata().get(0).getFieldTypes().get(fieldIdx).getName().equals("byte");
		targetBinary = getComponent().getOutMetadata().get(0).getFieldTypes().get(targetIdx).getName().equals("byte");
	}

	@Override
	public void preExecute() throws ComponentNotReadyException {
		super.preExecute();
	}

	@Override
	public void postExecute() throws ComponentNotReadyException {
		super.postExecute();
	}
	
	/**
	 * Interface for encryption algorithm wrapper
	 * @author Pavel Svec <pavel.svec@javlin.eu>
	 */
	private interface IEncryptionApi {
		/**
		 * Changes mode of an algorithm between encryption/decryption
		 * @param mode
		 * @throws InvalidKeyException
		 * @throws InvalidAlgorithmParameterException
		 */
		public void setMode(CipherMode mode) throws InvalidKeyException, InvalidAlgorithmParameterException;
		/**
		 * Encrypts/decrypts string passed as an argument
		 * @param in
		 * @return
		 */
		public byte[] encrypt(byte[] in) throws ShortBufferException, ShortBufferException, IllegalBlockSizeException, BadPaddingException, IllegalStateException;
	}
	
	/**
	 * Wrapper for Bouncy Castle AES block cipher algorithm
	 * @author Pavel Svec <pavel.svec@javlin.eu>
	 */
	private class AesEncryptor implements IEncryptionApi {
		
		Cipher cipher;
		IvParameterSpec ivSpec;
		Key key;
		
		public AesEncryptor(byte[] initialVector, byte[] key) throws NoSuchAlgorithmException, NoSuchProviderException, NoSuchPaddingException, InvalidKeyException {
			this.key = new SecretKeySpec(key, "DES");
	        ivSpec = new IvParameterSpec(initialVector);
	        cipher = Cipher.getInstance("AES/CTR/NoPadding", "BC");
		}
		
		public void setMode(CipherMode mode) throws InvalidKeyException, InvalidAlgorithmParameterException {
			switch (mode) {
			case ENCRYPT:
				cipher.init(Cipher.ENCRYPT_MODE, key, ivSpec);
				break;
			case DECRYPT:
				cipher.init(Cipher.DECRYPT_MODE, key, ivSpec);
				break;
			}	
		}
		
		public byte[] encrypt(byte[] in) throws ShortBufferException, ShortBufferException, IllegalBlockSizeException, BadPaddingException, IllegalStateException {
			byte[] cipherText;
			cipherText = new byte[cipher.getOutputSize(in.length)];
			
			int ctLength = cipher.update(in, 0, in.length, cipherText, 0);
			cipher.doFinal(cipherText,ctLength);
			
			return cipherText;
		}
	}
	
	/**
	 * Wrapper for Bouncy Castle ARC4 stream cipher algorithm
	 * @author Pavel Svec <pavel.svec@javlin.eu>
	 */
	private class Arc4Encryptor implements IEncryptionApi {
		
		Cipher cipher;
		SecretKeySpec key;
		
		public Arc4Encryptor(byte[] key) throws NoSuchAlgorithmException, NoSuchProviderException, NoSuchPaddingException {
			this.key = new SecretKeySpec(key, "ARC4");
			cipher = Cipher.getInstance("ARC4", "BC");
		}
		
		@Override
		public void setMode(CipherMode mode) throws InvalidKeyException, InvalidAlgorithmParameterException {
			switch (mode) {
			case ENCRYPT:
				cipher.init(Cipher.ENCRYPT_MODE, key);
				break;
			case DECRYPT:
				cipher.init(Cipher.DECRYPT_MODE, key);
				break;
			}	
		}

		@Override
		public byte[] encrypt(byte[] in) throws ShortBufferException, ShortBufferException, IllegalBlockSizeException, BadPaddingException, IllegalStateException {
			byte[] cipherText = new byte[in.length];
			int ctLength = cipher.update(in, 0, in.length, cipherText, 0);
			
			ctLength += cipher.doFinal(cipherText, ctLength);
			return cipherText;
		}
		
	}
	
	/**
     * Returns byte array of hex string.
     * 
     * @param s
     * @return
     */
    public static byte[] fromHex(String hex) {
        int bytes = hex.length() / 2;
        
        byte[] arr = new byte[bytes];
        for (int i = 0; i < bytes; i++) {
            arr[i] = (byte) ((Character.digit(hex.charAt(i*2), 16) << 4) + Character.digit(hex.charAt(i*2+1), 16));
        }
        
        return arr;
    }
	
	/*
	 * =========================================================================
	 * Following methods were taken from Bouncy Castle example projects "as are"
	 * =========================================================================
	 */
	
	/**
     * Convert the passed in String to a byte array by
     * taking the bottom 8 bits of each character it contains.
     * 
     * @param string the string to be converted
     * @return a byte array representation
     */
    public static byte[] toByteArray(
        String string)
    {
        byte[]	bytes = new byte[string.length()];
        char[]  chars = string.toCharArray();
        
        for (int i = 0; i != chars.length; i++)
        {
            bytes[i] = (byte)chars[i];
        }
        
        return bytes;
    }
    
    /**
     * Return length many bytes of the passed in byte array as a hex string.
     * 
     * @param data the bytes to be converted.
     * @param length the number of bytes in the data block to be converted.
     * @return a hex representation of length bytes of data.
     */
    public static String toHex(byte[] data, int length)
    {
    	final String digits = "0123456789ABCDEF";
        StringBuffer	buf = new StringBuffer();
        
        for (int i = 0; i != length; i++)
        {
            int	v = data[i] & 0xff;
            
            buf.append(digits.charAt(v >> 4));
            buf.append(digits.charAt(v & 0xf));
        }
        
        return buf.toString();
    }
    
    /**
     * Return the passed in byte array as a hex string.
     * 
     * @param data the bytes to be converted.
     * @return a hex representation of data.
     */
    public static String toHex(byte[] data)
    {
        return toHex(data, data.length);
    }
}
